(def
 #^{:private true}
 *special-form-docs*
{

'&
"-------------------------
& rest-param
Special Form
  Introduces rest-param in a parameter list. See fn*."

'.
"-------------------------
(. instance-expr member-symbol)
(. Classname-symbol member-symbol)
(. instance-expr (method-symbol arg*))
(. instance-expr method-symbol arg*)
(. Classname-symbol (method-symbol arg*))
(. Classname-symbol method-symbol arg*)
Special Form
  The '.' special form is the primary access to Java. It can be considered
  a member-access operator, and/or read as 'in the scope of'.

  If the first operand is a symbol that resolves to a class name, the
  access is considered to be to a static member of the named
  class. Otherwise it is presumed to be an instance member and the first
  argument is evaluated to produce the target object.

  If the second operand is a symbol and no args are supplied it is taken to
  be a field access - the name of the field is the name of the symbol, and
  the value of the expression is the value of the field, unless there is a
  no argument public method of the same name, in which case it resolves to
  a call to the method.

  If the second operand is a list, or args are supplied, it is taken to be
  a method call. The first element of the list must be a simple symbol, and
  the name of the method is the name of the symbol. The args, if any, are
  evaluated from left to right, and passed to the matching method, which is
  called, and its value returned. If the method has a void return type, the
  value of the expression will be nil. Note that placing the method name in
  a list with any args is optional in the canonic form, but can be useful
  to gather args in macros built upon the form.

  Note that boolean return values will be turned into Booleans, chars will
  become Characters, and numeric primitives will become Numbers unless they
  are immediately consumed by a method taking a primitive."

'catch
"-------------------------
(catch classname name expr*)
Special Form
  Introduces a catch clause. See try."

'def
"-------------------------
(def symbol init?)
Special Form
  Creates and interns or locates a global var with the name of symbol and a
  namespace of the value of the current namespace (*ns*). If init is
  supplied, it is evaluated, and the root binding of the var is set to the
  resulting value. If init is not supplied, the root binding of the var is
  unaffected. def always applies to the root binding, even if the var is
  thread-bound at the point where def is called. def yields the var
  itself (not its value). Throws an exception if symbol is already in the
  namespace and not mapped to an interned var.

  Any metadata on the symbol will be evaluated, and become metadata on the
  var itself. There are several metadata keys that have special
  interpretation:

    * :private
      a boolean indicating the access control for the var. If this key is
      not present, the default access is public (e.g. as if :private
      false).

    * :doc
      a string containing short (1-3 line) documentation for the var
      contents

    * :test
      a fn of no args that uses assert to check various operations. The var
      itself will be accessible during evaluation of a literal fn in the
      metadata map.

    * :tag
      a symbol naming a class or a Class object that indicates the Java
      type of the object in the var, or its return value if the object is a
      fn.

  In addition the compiler will place the following keys on the var:

    * :file string
    * :line int
    * :name simple symbol
    * :ns namespace in which var is interned
    * :macro true if var names a macro
    * :arglists a list of vector(s) of argument forms, as were supplied to
      defn

  The var metadata can be used for application-specific purposes as
  well. Consider using namespace-qualified keys (e.g. :myns/foo) to avoid
  clashes.

  (defn
    #^{:doc \"mymax [xs+] gets the maximum value in xs using > \"
       :test (fn []
               (assert (= 42  (max 2 42 5 4))))
       :user/comment \"this is the best fn ever!\"}
    mymax
    ([x] x)
    ([x y] (if (> x y) x y))
    ([x y & more]
       (reduce mymax (mymax x y) more)))

  user=> ^#'mymax
  {:name mymax
   :user/comment \"this is the best fn ever!\"
   :doc \"mymax [xs+] gets the maximum value in xs using >\"
   :file \"REPL\"
   :ns #
   :test user.fn__7290@2c3438
   :line 126}

  Many macros expand into def (e.g. defn, defmacro), and thus also convey
  metadata for the resulting var from the symbol used as the name.

  Using def to modify the root value of a var at other than the top level
  is usually an indication that you are using the var as a mutable global
  and is considered bad style. Consider either using binding to provide a
  thread-local value for the var, or putting a ref or agent in the var and
  using transactions or actions for mutation."

'do
"-------------------------
(do expr*)
Special Form
  Evaluates the expressions in order and returns the value of the last. If
  no expressions are supplied, returns nil."

'finally
"-------------------------
(finally expr*)
Special Form
  Introduces a finally clause. See try."

'fn*
"-------------------------
(fn* name? [param*] expr*)
(fn* name? ([param*] expr*)+)
params => positional-param* , or positional-param* & rest-param
positional-param => symbol
rest-param => symbol
name => symbol
Special Form
  Defines a function (fn). Fns are first-class objects that implement the
  IFn interface. The IFn interface defines an invoke() function that is
  overloaded with arity ranging from 0-20. A single fn object can implement
  one or more invoke methods, and thus be overloaded on arity. One and only
  one overload can itself be variadic, by specifying the ampersand followed
  by a single rest-param. Such a variadic entry point, when called with
  arguments that exceed the positional params, will find them in a seq
  contained in the rest param. If the supplied args do not exceed the
  positional params, the rest param will be nil.

  The first form defines a fn with a single invoke method. The second
  defines a fn with one or more overloaded invoke methods. The arities of
  the overloads must be distinct. In either case, the result of the
  expression is a single fn object.

  The exprs are compiled in an environment in which the params are bound to
  the actual arguments. The exprs are enclosed in an implicit do. If a name
  symbol is provided, it is bound within the function definition to the
  function object itself, allowing for self-calling, even in anonymous
  functions. If a param symbol is annotated with a metadata tag, the
  compiler will try to resolve the tag to a class name and presume that
  type in subsequent references to the binding.

  (def mult
       (fn this
         ([] 1)
         ([x] x)
         ([x y] (. clojure.lang.Num (multiply x y)))
         ([x y & more]
            (apply this (this x y) more))))

  Note that named fns such as mult are normally defined with defn, which
  expands into something such as the above.

  A fn (overload) defines a recursion point at the top of the function
  with arity equal to the number of params including the rest param, if
  present. See recur.

  IFns implement the Java Callable, Runnable and Comparator interfaces."

'if
"-------------------------
(if test then)
(if test then else)
Special Form
  Evaluates test. If not nil or false, evaluates and yields then
  otherwise, evaluates and yields else. If else is not supplied it defaults
  to nil. All of the other conditionals in Clojure are based upon the same
  logic, that is, nil and false constitute logical falsity, and everything
  else constitutes logical truth, and those meaning apply throughout. if
  performs conditional test of boolean Java method returns without
  conversion to Boolean."

'let*
"-------------------------
(let* [binding*] expr*)
binding => symbol init-expr
Special Form
  Evaluates the exprs in a lexical context in which the symbols in the
  binding-forms are bound to their respective init-exprs. The bindings are
  sequential, so each binding can see the prior bindings. The exprs are
  contained in an implicit do. If a binding symbol is annotated with a
  metadata tag, the compiler will try to resolve the tag to a class name
  and presume that type in subsequent references to the binding."

'loop*
"-------------------------
(loop* [binding*] expr*)
binding => symbol init-expr
Special Form
  loop is exactly like let, except that it establishes a recursion point at
  the top of the loop, with arity equal to the number of bindings. See
  recur."

'monitor-enter
"-------------------------
(monitor-enter x)
Special Form
  Synchronization primitive that should be avoided in user code. Use the
  locking macro."

'monitor-exit
"-------------------------
(monitor-enter x)
Special Form
  Synchronization primitive that should be avoided in user code. Use the
  locking macro."

'new
"-------------------------
(new Classname-symbol arg*)
Special Form
  The args, if any, are evaluated from left to right, and passed to the
  constructor of the class named by the symbol. The constructed object is
  returned."

'quote
"-------------------------
(quote form)
Special Form
  Yields the unevaluated form. The reader macro 'x expands to (quote x)."

'recur
"-------------------------
(recur expr*)
Special Form
  Evaluates the exprs in order, then, in parallel, rebinds the bindings of
  the recursion point to the values of the exprs. If the recursion point
  was a fn method, then it rebinds the params. If the recursion point was a
  loop, then it rebinds the loop bindings. Execution then jumps back to the
  recursion point. The recur expression must match the arity of the
  recursion point exactly. In particular, if the recursion point was the
  top of a variadic fn method, there is no gathering of rest args - a
  single seq (or null) should be passed. recur in other than a tail
  position is an error.

  Note that recur is the only non-stack-consuming looping construct in
  Clojure. There is no tail-call optimization and the use of self-calls for
  looping of unknown bounds is discouraged. recur is functional and its use
  in tail-position is verified by the compiler."

'set!
"-------------------------
(set! (. instance-expr instanceFieldName-symbol) expr)
(set! (. Classname-symbol staticFieldName-symbol) expr)
(set! var-symbol expr)
Special Form
  Assignment special form.

  When the first operand is a field member access form, the assignment is
  to the corresponding field. If it is an instance field, the instance expr
  will be evaluated, then the expr.

  When the first operand is a symbol, it must resolve to a global var. The
  value of the var's current thread binding is set to the value of
  expr. Currently, it is an error to attempt to set the root binding of a
  var using set!, i.e. var assignments are thread-local.

  In all cases the value of expr is returned.

  Note - you cannot assign to function params or local bindings. Only Java
  fields, Vars, Refs and Agents are mutable in Clojure."

'throw
"-------------------------
(throw expr)
Special Form
  The expr is evaluated and thrown, therefor it should yield an instance of
  some derivee of Throwable."

'try
"-------------------------
(try expr* catch-clause* finally-clause?)
catch-clause => (catch classname name expr*)
finally-clause => (finally expr*)
Special Form
  The exprs are evaluated and, if no exceptions occur, the value of the
  last is returned. If an exception occurs and catch clauses are provided
  each is examined in turn and the first for which the thrown exception is
  an instance of the named class is considered a matching catch clause. If
  there is a matching catch clause, its exprs are evaluated in a context in
  which name is bound to the thrown exception, and the value of the last is
  the return value of the function. If there is no matching catch clause
  the exception propagates out of the function. Before returning, normally
  or abnormally, any finally exprs will be evaluated for their side
  effects."

'var
"-------------------------
(var symbol)
Special Form
  The symbol must resolve to a var, and the Var object itself (not its
  value) is returned. The reader macro #'x expands to (var x)."

})

(defn print-special-form-doc
  [x]
  (println (*special-form-docs* x)))
